(require (lib "list.ss"))
;;build an empty sudoku board (with no helpful constraints)
;;
(define (make-empty-board)
  (let* ((cells   (make-cells))
         (rows    (add-cells-to-rows cells))
         (cols    (add-cells-to-cols cells))
         (squares (add-cells-to-squares cells)))
    (link-cells-to-rows rows )
    (link-cells-to-cols cols )
    (link-cells-to-squares squares )
    ;;maybe link rows,squares and cols,squares? 
    ;;not necessary but students may want them 
    ;;for extra-nifty implementations
    (list 'board cells rows cols squares)))
  
(define (make-cells)
  (let loop ((cells '())
             (iterator 0))
    (if (> 81 iterator)
        (loop (cons (list 'cell iterator (list 'values 1 2 3 4 5 6 7 8 9) 
                          (list 'row-num) (list 'col-num) (list 'square-num)) cells) (+ 1 iterator))
        (cons 'cells cells))))
(define (add-cells-to-rows cells)
  (let loop1 ((rows '())
             (iterator1 0))
    (let loop2 ((row '())
                (iterator2 0))
      (if (> 9 iterator2)
          (let ((cell (findf 
                       (lambda (c) (eq? (second c) (+(* iterator1 9)iterator2))) 
                       (rest cells))))
            (loop2 (cons cell row) (+ iterator2 1)))
          (if (> 9 iterator1)
              (loop1 (cons (append (list 'row iterator1) row) rows)(+ 1 iterator1))
              (cons 'rows rows))))))
(define (add-cells-to-cols cells)
  (let loop1 ((cols '())
             (iterator1 0))
    (let loop2 ((col '())
                (iterator2 0))
      (if (> 9 iterator2)
          (let ((cell (findf 
                       (lambda (c) (eq? (second c) (+(* iterator2 9)iterator1))) 
                       (rest cells))))
            (loop2 (cons cell col) (+ iterator2 1)))
          (if (> 9 iterator1)
              (loop1 (cons (append (list 'col iterator1) col) cols)(+ 1 iterator1))
              (cons 'cols cols))))))
(define (add-cells-to-squares cells)
  (let loop1 ((squares '())
             (iterator1 0))
    (let loop2 ((square '())
                (iterator2 0))
      (if (> 9 iterator2)
          (let ((cell (findf 
                       (lambda (c) (eq? (second c) (+ (modulo iterator2 3)
                                                      (*(floor(/ iterator2 3)) 9)
                                                      (modulo (* iterator1 3) 9)
                                                      (*(floor(/ iterator1 3))27))))                           
                       (rest cells))))
            (loop2 (cons cell square) (+ iterator2 1)))
          (if (> 9 iterator1)
              (loop1 (cons (append (list 'square iterator1) square) squares)(+ 1 iterator1))
              (cons 'squares squares))))))
(define (link-cells-to-rows rows)
  (let loop1 ((rest-rows (cdr rows)))
    (let loop2 ((rest-cells (rest (rest(car rest-rows)))))
      (append! (fourth (first rest-cells)) (list (row-number (car rest-rows))))
      (if (null? (cdr rest-cells))
          (if (null? (cdr rest-rows))
              #t
              (loop1 (rest rest-rows)))
          (loop2 (rest rest-cells)))
      
     )))
(define (link-cells-to-cols cols)
  (let loop1 ((rest-cols (cdr cols)))
    (let loop2 ((rest-cells (rest (rest(car rest-cols)))))
      (append! (fifth (first rest-cells)) (list (col-number (car rest-cols))))
      (if (null? (cdr rest-cells))
          (if (null? (cdr rest-cols))
              #t
              (loop1 (rest rest-cols)))
          (loop2 (rest rest-cells)))
      
     )))
(define (link-cells-to-squares squares)
  (let loop1 ((rest-squares (cdr squares)))
    (let loop2 ((rest-cells (rest (rest(car rest-squares)))))
      (append! (sixth (first rest-cells)) (list (square-number (car rest-squares))))
      (if (null? (cdr rest-cells))
          (if (null? (cdr rest-squares))
              #t
              (loop1 (rest rest-squares)))
          (loop2 (rest rest-cells)))
      
     )))
;;get row/col/square/cell by number:
(define (get-x-by-number x-list number)
  (findf (lambda (x) (eq? (second x) number)) x-list))
;;board accesssor methods:
(define (get-rows board)    (rest(third board)))
(define (get-cols board)    (rest(fourth board)))
(define (get-squares board) (rest(fifth board)))
(define (get-cells board)   (rest(second board)))
(define (get-row-by-number board row-num) 
  (if (or(< row-num 0)(> row-num 8)) (error "row number out of range"))
  (get-x-by-number (get-rows board) row-num))
(define (get-col-by-number board col-num) 
  (if (or(< col-num 0)(> col-num 8)) (error "col number out of range"))
  (get-x-by-number (get-cols board) col-num))
(define (get-square-by-number board square-num) 
  (if (or(< square-num 0)(> square-num 8)) (error "square number out of range"))
  (get-x-by-number (get-squares board) square-num))
(define (get-cell-by-number board cell-num) 
  (if (or(< cell-num 0)(> cell-num 80)) (error "cell number out of range"))
  (get-x-by-number (get-cells board) cell-num))
;;cell methods:
(define (cell-number cell) (second cell))
(define (cell-values cell) (rest(third cell))) ;; the current set of value assignments
(define (cell-row-number cell)    (second (fourth cell)))
(define (cell-col-number cell)    (second (fifth cell)))
(define (cell-square-number cell) (second (sixth cell)))
(define (cells-in-same-row cell board) 
  (remove cell (cells-in-row (get-row-by-number board (cell-row-number cell)))))
(define (cells-in-same-col cell board) 
  (remove cell (cells-in-col (get-col-by-number board (cell-col-number cell)))))
(define (cells-in-same-square cell board) 
  (remove cell (cells-in-square (get-square-by-number board (cell-square-number cell)))))
(define (set-values! cell new-values)
  (set-cdr! (third cell) new-values))
  
;;row methods:
(define (row-number row) (second row))
(define (cells-in-row row) (cddr row))
;;col methods:
(define (col-number col) (second col))
(define (cells-in-col col) (cddr col))
;;square methods:
(define (square-number square) (second square))
(define (cells-in-square square) (cddr square))

;;turn a board into something we can use for the tester:
(define (summarize-board board)
  (let* ((cells (get-cells board))
         (scells (sort cells (lambda (c1 c2)(< (cell-number c1)(cell-number c2)))))
         (vals (map cell-values scells)))
    vals))
             
(define (fix-cell-num num val board)
  (let ((cell (get-cell-by-number board num)))
    (set-values! cell (list val))
    (append! cell (list 'given))))

;;build the board used in the pset
(define (testing-board)
  (define board (make-empty-board))
  (fix-cell-num 1 2 board)
  (fix-cell-num 2 8 board)
  (fix-cell-num 3 6 board)
  (fix-cell-num 5 3 board)
  (fix-cell-num 6 4 board)
  (fix-cell-num 10 3 board)
  (fix-cell-num 12 7 board)
  (fix-cell-num 14 4 board)
  (fix-cell-num 15 5 board)
  (fix-cell-num 16 8 board)
  (fix-cell-num 20 1 board)
  (fix-cell-num 24 6 board)
  (fix-cell-num 30 9 board)
  (fix-cell-num 31 3 board)
  (fix-cell-num 32 6 board)
  (fix-cell-num 34 1 board)
  (fix-cell-num 35 4 board)
  (fix-cell-num 37 6 board)
  (fix-cell-num 38 3 board)
  (fix-cell-num 39 1 board)
  (fix-cell-num 41 7 board)
  (fix-cell-num 44 5 board)
  (fix-cell-num 46 7 board)
  (fix-cell-num 49 5 board)
  (fix-cell-num 51 9 board)
  (fix-cell-num 52 3 board)
  (fix-cell-num 53 6 board)
  (fix-cell-num 57 5 board)
  (fix-cell-num 58 8 board)
  (fix-cell-num 63 3 board)
  (fix-cell-num 65 6 board)
  (fix-cell-num 67 7 board)
  (fix-cell-num 69 2 board)
  (fix-cell-num 73 8 board)
  (fix-cell-num 74 5 board)
  (fix-cell-num 75 3 board)
  board)
