(require (lib "list.ss"))
(load "utils.scm")
(load "data.scm")

;; PROBLEM 1

;; 1.1.1
;; Returns the hamming distance between two feature sets
(define (hamming-distance x y)
  (+ (if (equal? (car x) (car y)) 0 1) (if (and (pair? (cdr x)) (pair? (cdr y))) (hamming-distance (cdr x) (cdr y)) 0))
)


;; 1.1.2
;; Returns the euclidean distance between two feature sets
(define (euclidean-distance x y)
  (sqrt (foldl + 0 (map (lambda (x1 y1) (expt (- x1 y1) 2)) x y)))
)

;; 1.1.3
;; Returns the cos(theta) between two feature vectors
(define (cos-theta x y)
  (/ (dot-product x y) (* (magnitude x) (magnitude y)))
)
  
(define (dot-product x y) (foldl + 0 (map * x y)))

(define (magnitude x) (sqrt (foldl + 0 (map (lambda (x1) (expt x1 2)) x))))


;; 1.2
;; Creates a nearest-neighbor classifier using the provided distance metric
;;  and training examples
;; Returns a procedure that given a set of features, returns a classification
(define (make-nn-classifier distance-metric training-examples)
  (lambda (features)
    (letrec ((nearest (lambda (examples min-class min-dist)
                       (if (not (pair? examples)) 
                           min-class
                           (let ((distance (distance-metric features (get-features (car examples)))))
                             (nearest (cdr examples) (if (< distance min-dist) (get-class (car examples)) min-class) (if (< distance min-dist) distance min-dist)) 
                           )
                      ))))
      (nearest (cdr training-examples) (get-class (car training-examples)) (distance-metric features (get-features (car training-examples))))
    )                   
  )
)


 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;; Hash-table usage example ;;;;
 ;; Create a new hashtable using the equal? operator (so strings can be compared)
 ;(define example-table (make-hash-table 'equal))
 ;; Store values for the keys "x" and "y":
 ;(hash-table-put! example-table "x" 5)
 ;(hash-table-put! example-table "y" 7)
 ;; Retrieve the value stored for the key "x".
 ;; The 0 is the default value to return if the key is not found:
 ;(hash-table-get example-table "x" 0)
 ;(hash-table-get example-table "z" 0)
 ;; Apply a procedure for every key-value pair:
 ;(hash-table-for-each example-table (lambda (key value) (print value)))
 ;; Or:
 ;(define key-list (hash-table-map example-table (lambda (key value) key)))
 ;(newline)
 ;(print key-list)
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; 1.3
;; Creates a k-nearest-neighbors classifier using the provided distance metric
;;  and training examples
;; Returns a procedure that given a set of features, returns a classification
(define (make-k-nn-classifier k distance-metric training-examples)
  (lambda (features)
    (let ([k-nearest-examples (get-k-nearest k distance-metric features training-examples)]
          [class-count-table (make-hash-table 'equal)])
      ;; Return the class most represented in the k nearest neighbors
      (let loop((examples k-nearest-examples))
        (hash-table-put! class-count-table (get-class (car examples)) (+ 1 (hash-table-get class-count-table (get-class (car examples)) 0)))
        (if (pair? (cdr examples)) (loop (cdr examples)))
      )
      (car (cdr (foldl (lambda (x1 x2) (if (> (car x1) (car x2)) x1 x2)) '(0 "BLANK") (append '((0 "BLANK")) (hash-table-map class-count-table (lambda (key value) (list value key)))))))
    )
  )
)


;; PROBLEM 2

;; 2.1
;; This is an elementary disorder calculation that merely negates the number
;; of elements in homogenous sets
(define (basic-disorder attribute-to-split classes examples)
 (- 0 (foldl + 0 
   (let loop((split (cadr attribute-to-split)) (features (caddr attribute-to-split)))    
    (cons (homogenous-size (group-by-attribute split (car features) examples))   
    (if (pair? (cdr features)) (loop split (cdr features)) '()))  
   )
 ))
)

;This returns the length if all the examples belong to the same class, 0 if otherwise
(define (homogenous-size examples)
  (if (pair? examples)
      (let ((class (get-class (car examples))))
        (if (foldr (lambda (x2 x1) (and x1 (equal? class (get-class x2)))) #t examples) (length examples) 0) 
        )
  0)
)

;This returns the set of examples which match class
(define (group-by-attribute split feature examples)
   (if (pair? examples)
     (if (equal? (split (get-features (car examples))) feature)
         (cons (car examples) (group-by-attribute split feature (cdr examples)))
         (group-by-attribute split feature (cdr examples)))
     '())
)

;(homogenous-size (group-by-attribute (cadr lotion-attribute-spec) (car (caddr lotion-attribute-spec)) sunburn-data)) 

;(basic-disorder height-attribute-spec '("sunburned" "none") sunburn-data)

;; PROBLEM 3

;; GENERAL NOTES
;;
;; A bpnn weight-list is a list of weight specifications.
;; A weight specification is a pair containing a weight matrix
;; for the layer as its car and a bias vector as its cdr.
;;
;; A weight matrix is a vector of rows; each row is a vector of
;; columns. An element at row i, column j in this matrix contains
;; the weight of the connection leading from node j in the previous
;; layer to node i in the current layer.
;;
;; Element i in the bias vector contains the weight of the connection
;; leading from the bias value (fixed to a constant 1) to node i in 
;; the current layer.
;;
;; The first element in the weight list is the symbol 'never-used; this
;; is because there are no weights leading into the input layer.
;;
;; The weight list is specified in this way so that you can use the procedures
;; in bpnn-helper.scm for doing basic linear algebraic computations (addition,
;; subtraction, elementwise-product) on vectors and matrices (e.g. matrix *
;; vector). Along with my-vector-map, this permits you to implement the formulas
;; for backpropagation in a concise and efficient manner. Some keywords to
;; search for are vector-add, vector-sub and vector-elementwise-product.
;;
;; If you aren't comfortable with matrix/vector notation, don't worry about it,
;; and just write loops implementing the formulas in an elementwise fashion.
;; Your code won't be evaluated on its speed.


;; 3.1
;;
;; Forward prop:
;; Takes a bpnn and returns a list of activations
;; (y0 y1 y2 y3 ... yn)
;; where y0 is the input vector
;; y1 is the activation of layer 1
;; etc
;;
;; yi is oi in Winston's notation
;;
;; sig is the sigmoidal function you'll be using for this part (tanh, in fact).
(define (forward-prop bpnn input-vec)
  (let ([weight-list (get-weight-list bpnn)]
        [sig (get-sigmoid bpnn)]
        [num-layers (get-num-layers bpnn)])
    'fill-me-in))


;; 3.2
;;Takes a bpnn and returns a list of errors (the delta terms).
;;The first element can be anything, as there's no error in the input layer
;;by definition. Use sigderiv evaluated at the activation for node k to
;;replace the o_k(1-o_k) term (as tanh' = 1-tanh^2, yielding a slightly 
;;different derivative than the function used in the text.)
(define (backward-prop bpnn act-list target-vec)
  (let ([weight-list (get-weight-list bpnn)]
        [sigderiv (get-sigmoid-derivative bpnn)]
        [num-layers (get-num-layers bpnn)])
    'fill-me-in))

(define how-many-hours-this-pset-took "")
(define what-i-found-interesting "")
(define what-i-found-boring "")